package com.imageresizer;

import static com.facebook.react.bridge.UiThreadUtil.runOnUiThread;

import android.content.Context;
import android.graphics.Bitmap;
import android.graphics.BitmapFactory;
import android.graphics.drawable.Drawable;
import android.util.Log;
import android.webkit.MimeTypeMap;

import androidx.annotation.NonNull;

import com.bumptech.glide.Glide;
import com.bumptech.glide.integration.webp.decoder.WebpDrawable;
import com.bumptech.glide.load.resource.gif.GifDrawable;
import com.facebook.react.bridge.Arguments;
import com.facebook.react.bridge.Promise;
import com.facebook.react.bridge.ReactApplicationContext;
import com.facebook.react.bridge.ReactContextBaseJavaModule;
import com.facebook.react.bridge.ReactMethod;
import com.facebook.react.bridge.ReadableArray;
import com.facebook.react.bridge.ReadableMap;
import com.facebook.react.bridge.WritableArray;
import com.facebook.react.bridge.WritableNativeArray;

import java.io.File;
import java.io.FileNotFoundException;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

public class ImageResizerModule extends ReactContextBaseJavaModule {
  public static String TAG = "ImageResizerModule";
  public static final String NAME = "ImageResizer";
  // Assuming you have a class variable/member
  private final ExecutorService executorService = Executors.newSingleThreadExecutor();

  private Context context;
  ImageResizerModule(ReactApplicationContext context) {
    super(context);
    this.context = getReactApplicationContext();
  }

  @Override
  @NonNull
  public String getName() {
    return NAME;
  }

  @Override
  public Map<String, Object> getConstants() {

    final Map<String, Object> constants = new HashMap<>();
    Map<String, String> scale = ScaleMode.asMap();
    Map<String, String> format = ImageFormatType.asMap();

    constants.put("Scale", scale);
    constants.put("Format", format);
    return constants;
  }

  @ReactMethod
  public void Static(String uri, ReadableMap userOptions, Promise promise) {
    StaticOptions options = new StaticOptions(userOptions);

    File file = new File(uri);
    if (!file.exists()) {
      promise.reject(TAG,"File does not exist");
    }

    // create bitmap
    Bitmap bitmap = BitmapFactory.decodeFile(uri);

    //  re-scale
    Bitmap newBitmap = BitmapUtility.resizeBitmap(bitmap, options);

    // create file
    try {
      File newFile = BitmapUtility.saveBitmapToFile(newBitmap, options.outFormat, options.quality);
      promise.resolve(newFile.getAbsolutePath());
    } catch (FileNotFoundException e) {
      promise.reject("Error in ResizeStaticImage:", e);
    }


  }

  @ReactMethod
  public void Animated(String uri, ReadableMap userOptions, Promise promise) {
    // validate options
    StaticOptions options = new StaticOptions(userOptions);

    // TODO check uri existence
    // Create output file
    File outputFile = new FileUtility().createOutputFile(options.extention);


    try {
      // 1. Get resource : resource = GifDrawable / WebpDrawable
      Object resource = getResource(uri);
      // 2. Get decoder : decoder MyGifDecoder / MyWebpDecoder
      MyDecoder decoder = DecoderFactory.getDecoder(resource);
      // 3. Get encoder : encoder MyGifEncoder / MyWebpEncoder
      MyEncoder encoder = EncoderFactory.getEncoder(outputFile, options);


      // 4. Start Encoding...\
//      boolean isBreak = false;
      for(int i=0; i<decoder.getFrameCount(); i++){
        Bitmap ogBitmap = decoder.getNextFrame();
        Bitmap bitmap = BitmapUtility.resizeBitmap(ogBitmap, options );

        ogBitmap.recycle();

        // if any fram is invalid then reject encoding
        if(bitmap == null){
          // isBreak = true; // continue; TODO ignore frames
          break;
        }

        encoder.addFrame(bitmap); // TODO add delay
        bitmap.recycle();
        //NEXT
        decoder.advance();
      }
      // 5. Close Encoder
      encoder.close();

      // 6. Return result
      promise.resolve(outputFile.getAbsolutePath());
    } catch (Exception e) {
      // 6. inform rejection
      promise.reject(TAG,e);
      if (outputFile.exists()) {
        if (outputFile.delete()) {
          Log.d(TAG, "File deleted successfully: " + outputFile.getAbsolutePath());
        } else {
          Log.w(TAG, "Failed to delete file: " + outputFile.getAbsolutePath());
        }
      }
    }


  }

  @ReactMethod
  public void StaticBatch(ReadableArray uris, ReadableMap userOptions, Promise promise) {
    StaticOptions options = new StaticOptions(userOptions);

    // Create an array to store the paths of resized images
    WritableArray resizedPaths = Arguments.createArray();

    try {

      for (int i = 0; i < uris.size(); i++) {
        String uri = uris.getString(i);

        File file = new File(uri);
        if (!file.exists()) continue;

        // create bitmap
        Bitmap bitmap = BitmapFactory.decodeFile(uri);
        if(bitmap == null) continue;

        // re-scale
        Bitmap newBitmap = BitmapUtility.resizeBitmap(bitmap, options);
        bitmap.recycle();
        if(newBitmap == null) continue;

        // create file
        File newFile = BitmapUtility.saveBitmapToFile(newBitmap, options.outFormat, options.quality);
        resizedPaths.pushString(newFile.getAbsolutePath());
      }

      promise.resolve(resizedPaths);
    } catch (Exception e) {
      promise.reject(TAG,"Error in ResizeStaticImagesBatch:", e);
    }
  }

  @ReactMethod
  public void AnimatedBatch(ReadableArray uris, ReadableMap userOptions, Promise promise) {

    StaticOptions options = new StaticOptions(userOptions);

    WritableArray successUris = new WritableNativeArray();

    try{
      // For each uri
      for(int i=0; i<uris.size(); i++){
        // Decode Encode
        String uri = processAll(uris.getString(i), options);
        if(uri!=null){
          successUris.pushString(uri);
        }
      }

      // resolve with uris
      promise.resolve(successUris);

    }catch (Exception e){
      Log.w(TAG,"e "+e.getMessage());
      // reject with error
      promise.reject(TAG,e);
    }


  }


  private String processAll(String uri, StaticOptions options) {

    // TODO check uri existence
    // Create output file
    File outputFile = new FileUtility().createOutputFile(options.extention);

    try {
      // 1. Get resource : resource = GifDrawable / WebpDrawable
      Object resource = getResource(uri);
      // 2. Get decoder : decoder MyGifDecoder / MyWebpDecoder
      MyDecoder decoder = DecoderFactory.getDecoder(resource);
      // 3. Get encoder : encoder MyGifEncoder / MyWebpEncoder
      MyEncoder encoder = EncoderFactory.getEncoder(outputFile, options);

      boolean isBroke = false;
      // 4. Start Encoding...
      for(int i=0; i<decoder.getFrameCount(); i++){
        Bitmap ogBitmap = decoder.getNextFrame();
        Bitmap bitmap = BitmapUtility.resizeBitmap(ogBitmap, options );

        ogBitmap.recycle();

        // if any fram is invalid then reject encoding
        if(bitmap == null){
          isBroke = true;
          break;
        }

        encoder.addFrame(bitmap); // TODO add delay
        bitmap.recycle();
        //NEXT
        decoder.advance();
      }
      // 5. Close Encoder
      encoder.close();

      // 6. Return result
      return outputFile.getAbsolutePath();

    } catch (Exception e) {
      // 6. inform rejection
      if (outputFile.exists()) {
        if (outputFile.delete()) {
          Log.d(TAG, "File deleted successfully: " + outputFile.getAbsolutePath());
        } else {
          Log.w(TAG, "Failed to delete file: " + outputFile.getAbsolutePath());
        }
      }
      return null;
    }
  }

  private Object getResource(String uri) throws ExecutionException, InterruptedException {
    String inExtention = MimeTypeMap.getFileExtensionFromUrl(uri);

    // TODO : resusability
    if(inExtention.equalsIgnoreCase(ImageFormatType.GIF.getType())){
      return Glide.with(this.context)
              .as(GifDrawable.class)
              .load(uri)
              .submit()
              .get();
    } else if(inExtention.equalsIgnoreCase(ImageFormatType.WEBP.getType())){
      return Glide.with(this.context)
              .as(WebpDrawable.class)
              .load(uri)
              .submit()
              .get();
    }

    return null;
  }


}
